"""
'   updater.py
'
'   Handles updating user habit data
"""


import pathlib
from datetime import datetime
from abc import ABC, abstractmethod

from habiter import __version__
from habiter.internal.utils.consts import HAB_DATE_FORMAT

from habiter.internal.utils.messenger import echo_info
from habiter.internal.file.operations import SQLiteDataFileOperations


class AbstractFileUpdater(ABC):
    """Abstract class that provides file updating behaviors"""

    def __init__(self, f_path: pathlib.Path) -> None:
        self.f_path = f_path

    @abstractmethod
    def update(self) -> None:
        """Abstract method that updates the contents of a file"""
        pass


class SQLiteDataFileUpdater(AbstractFileUpdater):
    def update(self) -> None:

        with SQLiteDataFileOperations() as fo:
            fo.cur.execute('SELECT * FROM meta_info')
            # There should be a single row from the meta_info table
            row = fo.cur.fetchone()

            logged_time = datetime.strptime(row['last_logged'],
                                            HAB_DATE_FORMAT).date()
            curr_time = datetime.now().date()

            # Check if habit data has already been updated
            if logged_time >= curr_time:
                return

            echo_info('Last accessed: {}\n'.format(row['last_logged']))

            # Update row
            fo.cur.execute('UPDATE meta_info SET version=?, '
                           'last_logged=? WHERE meta_id=?',
                           (__version__,
                            datetime.now().strftime(HAB_DATE_FORMAT),
                            row['meta_id']))
            # Retrieve all habit data with active column set to 'True'
            fo.cur.execute('SELECT curr_tally, prev_tally, is_active, '
                           'last_updated, num_of_trials, habit_name, habit_id '
                           'FROM habit WHERE is_active=True')
            data = fo.cur.fetchall()
            for habit in data:
                habit_date = datetime \
                    .strptime(habit["last_updated"], HAB_DATE_FORMAT) \
                    .date()
                # TODO: implementation really inefficient but sqlite3 is challenging to use
                if habit_date < curr_time:
                    prev_tally = 0
                    num_of_trials = habit['num_of_trials'] + 1
                    is_active = False
                    curr_tally = 0
                    fo.cur.execute('UPDATE habit SET '
                                   'prev_tally = ?, num_of_trials = ?, '
                                   'is_active = ?, curr_tally = ? '
                                   'WHERE habit_id = ?',
                                   (prev_tally,
                                    num_of_trials,
                                    is_active,
                                    curr_tally,
                                    habit['habit_id']))



import sys
import click
from datetime import datetime

from habiter.internal.utils.consts import HAB_DATE_FORMAT
from habiter.internal.utils.messenger import echo_success, echo_failure
from habiter.internal.file.operations import SQLiteDataFileOperations


@click.command(short_help='add new habit(s) into record')
@click.argument('habits', required=True, nargs=-1)
def add(habits):
    existing_habit_detected = False

    # Cast to set to remove possible duplicates
    habits = set(habits)

    with SQLiteDataFileOperations() as fo:

        # TODO: Instead of invoking a query on each habit_name, try to use one query for all habits
        for habit_name in habits:
            fo.cur.execute('SELECT curr_tally, prev_tally, is_active, '
                           'last_updated, num_of_trials FROM habit WHERE '
                           'habit_name=?', (habit_name,))
            row = fo.cur.fetchone()
            if row is not None:
                existing_habit_detected = True
                echo_failure(f"Habit \"{habit_name}\" already exists.")
                continue
            curr_time = datetime.now().strftime(HAB_DATE_FORMAT)
            fo.cur.execute(
                'INSERT INTO habit (habit_name, curr_tally, '
                'total_tally, num_of_trials, wait_period, is_active, '
                'last_updated, date_added) '
                'VALUES(?, 0, 0, 0, 0, False, ?, ?)',
                (habit_name, curr_time, curr_time))
            echo_success(f"Habit \"{habit_name}\" has been added.")

        if existing_habit_detected:
            sys.exit(1)



"""
'   math.py
'
'   Provides stat and other mathematical
'   functions for habiter to work with
'
"""

import math
import sys

from habiter.internal.utils.messenger import echo_failure


# TODO: Introduce more math-related logic into the system!

def neumaier_sum(summ, val):
    raise NotImplementedError


def avg_percent_delta(avg1: float, avg2: float, time=1):
    """ Computes the average percent change
        for some function of time, expressed
        as a percentage (though it returns a
        float value)
        So,
            avg2 - avg1      1
            -----------  *  ---- * 100
                avg1        time

        represents the computation found below

        If it inputs a '-': represents % decrease
        If it inputs a '+': represents % increase
    Parameters
        avg1:   the old average
        avg2:   the new average
        time:   an int value for a function of
                time (by default is one for 'one
                day')
    """
    try:
        return (((avg2 - avg1) / avg1) * (1 / time)) * 100
    except ZeroDivisionError:
        echo_failure("Division by zero.")
        sys.exit(1)


def avg(summation, nTrials: int):
    try:
        return summation / nTrials
    except ZeroDivisionError:
        echo_failure("Division by zero.")
        sys.exit(1)


# Unused function, needs to be tested
def running_avg(avg, newVal, nTrials):
    old_avg = avg * (nTrials - 1)
    return old_avg + ((newVal - old_avg) / nTrials)


def poisson_prob(lmda, x: int):
    """ Computes the probability for a discrete random
        variable using the Poisson probability
    Parameters
        lmda:   the lambda parameter
        x:      a discete event that is observed (i.e. for occurrence(s))
    """
    return (math.e) ** (-lmda) * ((lmda ** x) / math.factorial(x))


if __name__ == "__main__":
    # print(avg(summation=5, nTrials=2))
    print(avg_percent_delta(4, 8))



'''
    Module that handles habiter normal
    and error display messages on the
    console.

'''
import click


def echo_success(text: str):
    click.secho(f"[habiter]  {text}", fg='cyan', bold=True)


def echo_failure(text: str):
    click.secho(f"[habiter: error]  {text}", fg='red', bold=True)


def echo_internal_failure(text: str):
    click.secho(f"[habiter: internal_error]  {text}", fg='red', bold=True)


def echo_warning(text: str):
    click.secho(f"[habiter]  {text}", fg='yellow', bold=True)


def echo_info(text: str):
    click.secho(f"[habiter]  {text}", bold=True)


def inquire_choice(choice: str) -> bool:
    echo_info(f"Are you sure you want to {choice}? This cannot be undone.")
    ans = ''
    while True:
        ans = input("[Provide a y/n.]: ")

        if ans != "y" and ans != "n":
            echo_warning("Please try again.")
        else:
            return True if ans == 'y' else False



"""
This file contains implementations involving
the creation of files used to r/w data
"""

import pathlib
import json
import sqlite3
from datetime import datetime
from abc import ABC, abstractmethod

from habiter import __version__
from habiter.internal.utils.consts import HAB_DATE_FORMAT, HAB_JSON_IND


class AbstractFileCreator(ABC):
    """An abstract class that defines file creation behaviors"""

    def __init__(self, dir_path: str, f_name: str):
        self.dir_path = pathlib.Path(dir_path)
        self.f_name = f_name
        self.data_f_path = self.dir_path / self.f_name

    def get_data_f_path(self) -> pathlib.Path:
        return self.data_f_path

    def create(self) -> None:
        """Creates a file with a directory path that is also recursively created if needed
        """
        # Does the child directory exist
        if not self.dir_path.exists():
            self.dir_path.mkdir(parents=True)

        if not self.data_f_path.exists():
            self._init_file()

    @abstractmethod
    def _init_file(self) -> None:
        """Abstract method that creates and initializes the contents of a file
        """
        pass


class SQLiteDataFileCreator(AbstractFileCreator):
    def _init_file(self) -> None:
        con = sqlite3.connect(self.data_f_path)

        # Create META_INFO table
        con.execute('''
        CREATE TABLE meta_info
        (meta_id        INTEGER  PRIMARY KEY AUTOINCREMENT,
            version        TEXT             NOT NULL,
            last_logged    TEXT             NOT NULL
        )
        ''')
        # Create HABIT table
        con.execute('''
                CREATE TABLE habit
                (
                    habit_id       INTEGER  PRIMARY KEY AUTOINCREMENT,
                    habit_name     TEXT              NOT NULL,
                    curr_tally     INT               NOT NULL,
                    total_tally    INT               NOT NULL,
                    num_of_trials  INT               NOT NULL,
                    wait_period    INT               NULL,
                    is_active      BOOLEAN           NOT NULL,
                    last_updated   TEXT              NOT NULL,
                    date_added     TEXT              NOT NULL,
                    prev_tally     INT               NULL
                )
                ''')
        # Initialize META_INFO table
        con.execute('INSERT INTO meta_info(version, last_logged) '
                    'VALUES (?, ?)',
                    (__version__,
                     datetime.now().strftime(HAB_DATE_FORMAT)))
        con.commit()
        con.close()


class JSONDataFileCreator(AbstractFileCreator):
    """Concrete class that held the original creation logic for the habiter data file.

    This will most likely be removed in future iterations but will be kept
    in case configuration files are introduced and the logic can be utilized
    in a similar manner.
    """

    def _init_file(self, f_path: str) -> None:
        with open(f_path, 'w') as f:
            # Initialize JSON arrays to hold JSON objects
            initFileContents = {
                "util": {
                    "version": __version__,
                    "last_logged": datetime.now().strftime(HAB_DATE_FORMAT)
                },
                "habits": []
            }
            json.dump(initFileContents, f, indent=HAB_JSON_IND)




import csv
import os
from abc import ABC, abstractmethod

from habiter.internal.file.operations import SQLiteDataFileOperations
from habiter.internal.utils.consts import HAB_TRACE_FPATH


# TODO: Ensure this abstraction can be extended
class AbstractDataFileExport(ABC):

    def __init__(self, export_name: str, export_dir: str = os.getcwd()):
        self.export_name = export_name
        self.export_dir = export_dir

    @abstractmethod
    def export(self) -> None:
        """Export the habiter data file"""


class CSVExport(AbstractDataFileExport):

    def export(self) -> None:

        with SQLiteDataFileOperations(HAB_TRACE_FPATH) as fo:
            # Retrieve all rows from database
            fo.cur.execute('SELECT * FROM habit')

            to_dir_path = os.path.join(self.export_dir,
                                       f'{self.export_name}.csv')
            with open(to_dir_path, 'w') as csvfile:
                csv_writer = csv.writer(csvfile)

                # Retrieve columns
                # 'i[0]' because description returns a 7-tuple for each column
                csv_writer.writerow(i[0] for i in fo.cur.description)

                # Retrieve rows
                csv_writer.writerows(fo.cur)



#!/usr/bin/env python3

"""
'   The entry point of habiter
"""

from appdirs import user_data_dir

import habiter.internal.cli as cli
from habiter.internal.file.creator import SQLiteDataFileCreator
from habiter.internal.file.updater import SQLiteDataFileUpdater
from habiter.internal.file.operations import SQLiteDataFileOperations

from habiter.internal.utils.consts import (
    HAB_AUTHOR, HAB_NAME, HAB_F_RECORD_NAME
)


def main():
    dir_path = user_data_dir(HAB_NAME, HAB_AUTHOR)
    data_file_creator = SQLiteDataFileCreator(dir_path, HAB_F_RECORD_NAME)
    data_file_creator.create()
    data_file_path = data_file_creator.get_data_f_path()

    # Set the data file path that all classes using this can have access to
    SQLiteDataFileOperations().set_f_path(data_file_path)

    data_file_updater = SQLiteDataFileUpdater(data_file_path)
    data_file_updater.update()

    cli.habiter()


if __name__ == "__main__":
    main()



import setuptools


def get_version(rel_path: str) -> str:
    with open(rel_path, 'r') as f:
        for line in f:
            if line.startswith('__version__'):
                delim = '"' if '"' in line else "'"
                version = line.split(delim)[1]
                return version
        raise RuntimeError('Unable to find version string')


with open('README.md', 'r', encoding='utf-8') as f:
    long_description = f.read()

setuptools.setup(
    name="habiter",
    version=get_version('habiter/__init__.py'),
    description="Quantifiy and keep tabs on habits.",
    author="Kemal Zebari",
    url="https://github.com/kemzeb/habiter",
    license="MIT",
    keywords=(
        "productivity",
        "cli"
    ),
    long_description=long_description,
    long_description_content_type='text/markdown',
    packages=setuptools.find_packages(
        exclude="test"
    ),

    install_requires=(
        'appdirs',
        'click',
    ),

    entry_points={
        "console_scripts": [
            "habiter=habiter.internal.run:main"
        ]
    },
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3"
    ],
    python_requires=">=3.8"
)



from datetime import datetime
import sys

import click
import sqlite3

import habiter.internal.math as habmath
from habiter.internal.file.operations import SQLiteDataFileOperations
from habiter.internal.utils.consts import HAB_DATE_FORMAT
from habiter.internal.utils.messenger import echo_failure, echo_info


@click.command(short_help='list all habits on record')
@click.option('-v', '--verbose', is_flag=True, default=False)
@click.argument('habits', nargs=-1, required=False)
def list(habits, verbose):
    nonExistingHabitDetected = False
    data = []
    habits = set(habits)  # Cast to set to remove possible duplicates
    with SQLiteDataFileOperations() as fo:
        if len(habits) != 0:
            for habit_name in habits:
                fo.cur.execute('SELECT * FROM habit WHERE '
                               'habit_name=?', (habit_name,))
                row = fo.cur.fetchone()
                if row is None:
                    nonExistingHabitDetected = True
                    echo_failure(f"Habit \"{habit_name}\" does not exist.")
                    continue
                data.append(row)
        else:
            data = fo.cur.execute('SELECT * FROM habit').fetchall()
        meta_data = fo.cur.execute('SELECT last_logged FROM meta_info') \
            .fetchone()
    if not verbose:
        print("Habit\n-------------------")
        for habit in data:
            print(habit["habit_name"])
        print("-------------------")
    else:
        curr_day = datetime.strptime(meta_data["last_logged"], HAB_DATE_FORMAT).date()
        print("Habit + Attributes\t\t\tValue")
        print("-------------------\t\t\t-----")
        for habit in data:
            print_verbose(habit, curr_day)
        print("-------------------\t\t\t-----")
        echo_info("Note: More data captured = increased statistical accuracy!\n")
    if nonExistingHabitDetected:
        sys.exit(1)


def print_verbose(habit: sqlite3.Row, curr_day):
    habit_day = datetime.strptime(
        habit["last_updated"], HAB_DATE_FORMAT).date()
    # subtraction returns a timedelta
    delta_day = (curr_day - habit_day).days

    print('[{}]'.format(habit["habit_name"]))
    if habit['num_of_trials'] < 2:
        probInfo = '(More data required)'
    else:
        avg = habit['total_tally'] / habit['num_of_trials']
        prob = ((1 -
                 habmath.poisson_prob(avg, 0) -  # Prob. occuring never
                 habmath.poisson_prob(avg, 1)) * 100)  # Prob. occurring once
        probInfo = f'{prob:.3f}%'

    print(f"  | P(Occurrences > 1 today):\t\t{probInfo}")
    print("  | Today's daily tally:\t\t{}".format(habit['curr_tally']))
    print("  | Total tally:\t\t\t{}".format(habit['total_tally']))
    print("  | # of days captured:\t\t\t{}".format(habit['num_of_trials']))
    print(f"  | Last updated:\t\t\t{delta_day} day(s) ago")
    print("  | Date added:\t\t\t\t{}\n".format(habit["date_added"]))



import click

from habiter.internal.commands import (
    add, list, remove, reset, tally
)
from habiter import __version__


@click.group(
    help="Quantifies and keeps tabs on unwanted habits you have developed over time.",
    epilog='For more information, visit the code repository at https://github.com/kemzeb/habiter.')
@click.version_option(version=__version__)
def habiter():
    pass


habiter.add_command(add.add)
habiter.add_command(list.list)
habiter.add_command(remove.remove)
habiter.add_command(reset.reset)
habiter.add_command(tally.tally)



import sys
import click
from datetime import datetime

from ._utils import abort_if_false
from habiter.internal.utils.consts import HAB_DATE_FORMAT
from habiter.internal.utils.messenger import echo_success, echo_failure
from habiter.internal.file.operations import SQLiteDataFileOperations


@click.command(short_help='reset some habit(s) from record')
@click.argument('habits', required=True, nargs=-1)
@click.option('--yes', is_flag=True, callback=abort_if_false,
              expose_value=False,
              prompt='Are you sure you want to reset habit(s) from record?')
def reset(habits):
    non_existing_habit_detected = False

    # Cast to set to remove possible duplicates
    habits = set(habits)

    with SQLiteDataFileOperations() as fo:

        for habit_name in habits:
            fo.cur.execute('SELECT habit_id FROM habit WHERE habit_name=?',
                           (habit_name,))
            row = fo.cur.fetchone()

            if row is None:
                echo_failure(f"No habit with the name \"{habit_name}\".")
                non_existing_habit_detected = True
            else:
                fo.cur.execute('UPDATE habit SET curr_tally = 0, '
                               'total_tally = 0, num_of_trials = 0, '
                               'wait_period = 0, is_active = False, '
                               'last_updated = ?, prev_tally = NULL '
                               'WHERE habit_id=?',
                               (datetime.now().strftime(HAB_DATE_FORMAT),
                                row['habit_id']))
                echo_success(f"Habit \"{habit_name}\" has been reset.")

        if non_existing_habit_detected:
            sys.exit(1)



"""This file provides the file operation logic for the files habiter
utilizes (e.g. using the context manager to reduce boilerplate project-wide)
"""

import pathlib
import sqlite3
from abc import ABC, abstractmethod


class AbstractFileOperations(ABC):
    """Abstract base class for file operations.

    Instances of this class all share the __f_path static variable which can
    only be retrieved or set by the getter and setter class methods
    """
    __f_path: pathlib.Path = None

    @classmethod
    def get_f_path(cls) -> pathlib.Path:
        return cls.__f_path

    @classmethod
    def set_f_path(cls, f_path) -> None:
        cls.__f_path = f_path

    @abstractmethod
    def __enter__(self):
        pass

    @abstractmethod
    def __exit__(self, exc_type, exc_val, exc_tb):
        pass


class SQLiteDataFileOperations(AbstractFileOperations):
    """Concrete class that provides database file operations using SQLite
    """

    def __init__(self):
        self.con = None
        self.cur = None

    def __enter__(self):
        self.con = sqlite3.connect(SQLiteDataFileOperations.get_f_path())
        # Allow having returned rows to map column names to values using
        # sqlite3.Row
        self.con.row_factory = sqlite3.Row
        self.cur = self.con.cursor()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.con.commit()
        self.con.close()



import sys
import click
from datetime import datetime

from habiter.internal.utils.consts import HAB_DATE_FORMAT
from habiter.internal.utils.messenger import echo_failure, echo_success, echo_warning
from habiter.internal.file.operations import SQLiteDataFileOperations


@click.command(short_help='increment the number of occurrences for some habit(s)')
@click.argument('habits', required=True, nargs=-1)
@click.option('-n', '--num', default=1)
@click.option('-z', '--zero', is_flag=True)
def tally(habits, num, zero):
    flag = False

    # Cast to set to remove possible duplicates
    habits = set(habits)

    with SQLiteDataFileOperations() as fo:
        for habit_name in habits:
            fo.cur.execute('SELECT habit_id, curr_tally, total_tally '
                           'FROM habit WHERE habit_name=?',
                           (habit_name,))
            row = fo.cur.fetchone()
            if row:
                if zero:
                    if row['curr_tally'] > 0:
                        echo_failure(
                            f'Habit "{habit_name}" already contains occurrences.')
                        flag = True
                        continue
                    # If a number was passed, make sure to inform the end user
                    # that the number given will be ignored.
                    if num != 1:
                        echo_warning(
                            "--zero flag passed; ignoring --num value.")
                    num = 0

                prev_tally = row['curr_tally']
                curr_tally = row['curr_tally'] + num
                total_tally = row['total_tally'] + num
                is_active = True
                last_updated = datetime.now().strftime(HAB_DATE_FORMAT)

                fo.cur.execute('UPDATE habit SET curr_tally=?, '
                               'total_tally=?, is_active=?, '
                               'last_updated=?, prev_tally=? '
                               'WHERE habit_id = ?',
                               (curr_tally, total_tally, is_active,
                                last_updated, prev_tally, row['habit_id']))
                if zero:
                    echo_success(f'Habit \"{habit_name}\" marked as active.')
                else:
                    echo_success(f'Habit "{habit_name}" tally updated '
                                 f'from {prev_tally} to {curr_tally}.')
            else:
                echo_failure(f"No habit with the name \"{habit_name}\".")
                flag = True

        if flag:
            sys.exit(1)



import sys
import click

from ._utils import abort_if_false
from habiter.internal.utils.messenger import (
    echo_failure, echo_success
)
from habiter.internal.file.operations import SQLiteDataFileOperations


@click.command(short_help='delete habit(s) from record')
@click.argument('habits', required=True, nargs=-1)
@click.option('--yes', is_flag=True, callback=abort_if_false,
              expose_value=False,
              prompt='Are you sure you want to delete habit(s) from record?')
def remove(habits):
    non_existing_habit_detected = False

    # Cast to set to remove possible duplicates
    habits = set(habits)

    with SQLiteDataFileOperations() as fo:

        for habit_name in habits:
            fo.cur.execute('SELECT habit_id FROM habit WHERE habit_name=?',
                           (habit_name,))
            row = fo.cur.fetchone()

            if row is None:
                echo_failure(f"No habit with the name \"{habit_name}\".")
                non_existing_habit_detected = True
            else:
                fo.cur.execute('DELETE FROM habit WHERE habit_id=?',
                               (row['habit_id'],))
                echo_success(f"Habit \"{habit_name}\" has been deleted.")

        if non_existing_habit_detected:
            sys.exit(1)



import pytest
from click.testing import CliRunner

from habiter.internal.cli import habiter
from habiter.internal.file.operations import SQLiteDataFileOperations
from habiter.internal.file.creator import SQLiteDataFileCreator


@pytest.fixture
def temp_f_name() -> str:
    return 'test.db'


@pytest.fixture
def temp_data_dir_path(tmpdir) -> str:
    return tmpdir.mkdir('test')


@pytest.fixture
def get_init_data_f_path(temp_data_dir_path, temp_f_name) -> str:
    SQLiteDataFileCreator(temp_data_dir_path, temp_f_name).create()
    return temp_data_dir_path.join(temp_f_name)


@pytest.fixture
def setup(get_init_data_f_path):
    SQLiteDataFileOperations.set_f_path(get_init_data_f_path)
    return SQLiteDataFileOperations.get_f_path()


def test_add(setup):
    runner = CliRunner()

    # Normal behavior
    result = runner.invoke(habiter, ['add', 't0', 't1'])
    assert result.exit_code == 0
    with SQLiteDataFileOperations() as fo:

        fo.cur.execute('SELECT habit_name FROM habit')
        rows = fo.cur.fetchall()
        assert rows[0]['habit_name'] in ['t0', 't1']
        assert rows[1]['habit_name'] in ['t0', 't1']

    # Add habit that already exists on record
    runner.invoke(habiter, ['add', 't2'])
    result = runner.invoke(habiter, ['add', 't2'])
    assert result.exit_code == 1
    with SQLiteDataFileOperations() as fo:
        fo.cur.execute('SELECT habit_name FROM habit '
                       'WHERE habit_name = "t2"')
        rows = fo.cur.fetchall()
        assert len(rows) < 2

    # Abnormal behavior (habit name is duplicated)
    result = runner.invoke(habiter, ['add', 't3', 't3'])
    assert result.exit_code == 0
    with SQLiteDataFileOperations() as fo:
        fo.cur.execute('SELECT habit_name FROM habit '
                       'WHERE habit_name = "t3"')
        rows = fo.cur.fetchall()
        assert len(rows) == 1


def test_list(setup):
    runner = CliRunner()

    # Print with empty record
    result = runner.invoke(habiter, ['list'])
    assert result.exit_code == 0

    # Print with habits in record
    runner.invoke(habiter, ['add', 't0', 't1'])
    result = runner.invoke(habiter, ['list'])
    assert result.exit_code == 0

    # Use verbose option
    result = runner.invoke(habiter, ['list', '-v'])
    assert result.exit_code == 0
    result = runner.invoke(habiter, ['list', '--verbose'])
    assert result.exit_code == 0

    # Provide existing habits as positional arguments
    result = runner.invoke(habiter, ['list', 't0', 't1', '-v'])
    assert result.exit_code == 0
    assert '[t0]' in result.output

    # Provide non-existing habits as positional arguments
    result = runner.invoke(habiter, ['list', 't1', 't99', '-v'])
    assert result.exit_code == 1
    assert '[t1]' in result.output


def test_remove(setup):
    runner = CliRunner()
    # Remove habit from an empty record
    result = runner.invoke(habiter, ['remove', '--yes', 't0'])
    assert result.exit_code == 1

    # Remove habit that exists on record
    runner.invoke(habiter, ['add', 't0'])
    runner.invoke(habiter, ['remove', '--yes', 't0'])

    with SQLiteDataFileOperations() as fo:
        fo.cur.execute('SELECT habit_name FROM habit WHERE habit_name="t0"')
        rows = fo.cur.fetchall()
        assert len(rows) == 0


def test_tally(setup):
    runner = CliRunner()
    # Tally habit that does not exist on record
    result = runner.invoke(habiter, ['tally', 't0'])
    assert result.exit_code == 1

    # Tally habit that does exist on record
    runner.invoke(habiter, ['add', 't0'])
    runner.invoke(habiter, ['tally', 't0'])
    with SQLiteDataFileOperations() as fo:
        fo.cur.execute('SELECT curr_tally FROM habit WHERE habit_name="t0"')
        row = fo.cur.fetchone()
        assert row['curr_tally'] == 1

    # Tally existing habit using '-n/--num' option
    runner.invoke(habiter, ['add', 't1'])
    runner.invoke(habiter, ['tally', '-n', 2, 't1'])
    runner.invoke(habiter, ['tally', '--num', 2, 't1'])
    with SQLiteDataFileOperations() as fo:
        fo.cur.execute('SELECT curr_tally FROM habit WHERE habit_name="t1"')
        row = fo.cur.fetchone()
        assert row['curr_tally'] == 4

    # Mark habit as active using '-z/--zero' option
    runner.invoke(habiter, ['add', 't2'])
    runner.invoke(habiter, ['tally', '-z', 't2'])
    runner.invoke(habiter, ['tally', '--zero', 't2'])
    with SQLiteDataFileOperations() as fo:
        # When a habit is added, it will by default be considered inactive
        # until the end user makes a tally, so we are testing to see if this
        # option marks it as active
        fo.cur.execute('SELECT is_active FROM habit WHERE habit_name="t2"')
        row = fo.cur.fetchone()
        assert row['is_active']

    # Use the '-z/--zero' option' but the habit has already been active
    result = runner.invoke(habiter, ['tally', '-z', 't1'])
    assert result.exit_code == 1  # evaluates to a true expression


def test_reset(setup):
    runner = CliRunner()

    # Reset habit that does not exist on record
    result = runner.invoke(habiter, ['reset', '--yes', 't0'])
    assert result.exit_code == 1

    # Reset habit that exists on record
    runner.invoke(habiter, ['add', 't0'])
    with SQLiteDataFileOperations() as fo:
        # Reset habit that exists on record
        fo.cur.execute('UPDATE habit SET curr_tally = 5, '
                       'total_tally = 30, num_of_trials = 4, '
                       'wait_period = 1, is_active = True, '
                       'prev_tally = 1 '
                       'WHERE habit_name="t0"')
    runner.invoke(habiter, ['reset', '--yes', 't0'])
    with SQLiteDataFileOperations() as fo:
        fo.cur.execute('SELECT * FROM habit WHERE habit_name="t0"')
        row = fo.cur.fetchone()
    assert row['curr_tally'] == 0
    assert row['total_tally'] == 0
    assert row['num_of_trials'] == 0
    assert row['wait_period'] == 0
    assert row['is_active'] == 0
    assert row['prev_tally'] is None
